# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.rings.finite_rings
r"""
Frobenius isogenies of elliptic curves

Frobenius isogenies only exist in positive characteristic `p`. They
are given by `\pi_n:(x,y)\mapsto (x^{p^n},y^{p^n})`.

This class implements `\pi_n` for `n \geq 0`. Together with existing
tools for composing isogenies (see :class:`EllipticCurveHom_composite`),
we can therefore represent arbitrary inseparable isogenies in Sage.

EXAMPLES:

Constructing a Frobenius isogeny is straightforward::

    sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
    sage: z5, = GF(17^5).gens()
    sage: E = EllipticCurve([z5,1])
    sage: pi = EllipticCurveHom_frobenius(E); pi
    Frobenius isogeny of degree 17:
      From: Elliptic Curve defined by y^2 = x^3 + z5*x + 1
            over Finite Field in z5 of size 17^5
      To:   Elliptic Curve defined by y^2 = x^3 + (9*z5^4+7*z5^3+10*z5^2+z5+14)*x + 1
            over Finite Field in z5 of size 17^5

By passing `n`, we can also construct higher-power Frobenius maps,
such as the Frobenius *endo*\morphism::

    sage: z5, = GF(7^5).gens()
    sage: E = EllipticCurve([z5,1])
    sage: pi = EllipticCurveHom_frobenius(E, 5); pi
    Frobenius endomorphism of degree 16807 = 7^5:
      From: Elliptic Curve defined by y^2 = x^3 + z5*x + 1
            over Finite Field in z5 of size 7^5
      To:   Elliptic Curve defined by y^2 = x^3 + z5*x + 1
            over Finite Field in z5 of size 7^5

The usual :class:`EllipticCurveHom` methods are supported::

    sage: z5, = GF(7^5).gens()
    sage: E = EllipticCurve([z5,1])
    sage: pi = EllipticCurveHom_frobenius(E,5)
    sage: pi.degree()
    16807
    sage: pi.rational_maps()
    (x^16807, y^16807)
    sage: pi.formal()                   # known bug
    ...
    sage: pi.is_normalized()            # known bug
    ...
    sage: pi.is_separable()
    False
    sage: pi.is_injective()
    True
    sage: pi.is_surjective()
    True

Computing the dual of Frobenius is supported as well::

    sage: E = EllipticCurve([GF(17^6).gen(), 0])
    sage: pi = EllipticCurveHom_frobenius(E)
    sage: pihat = pi.dual(); pihat
    Isogeny of degree 17
     from Elliptic Curve defined by y^2 = x^3 + (15*z6^5+5*z6^4+8*z6^3+12*z6^2+11*z6+7)*x
          over Finite Field in z6 of size 17^6
       to Elliptic Curve defined by y^2 = x^3 + z6*x
          over Finite Field in z6 of size 17^6
    sage: pihat.is_separable()
    True
    sage: pihat * pi == EllipticCurveHom_scalar(E,17)   # known bug -- #6413
    True

A supersingular example (with purely inseparable dual)::

    sage: E = EllipticCurve([0, GF(17^6).gen()])
    sage: E.is_supersingular()
    True
    sage: pi1 = EllipticCurveHom_frobenius(E)
    sage: pi1hat = pi1.dual(); pi1hat
    Composite morphism of degree 17 = 17*1:
      From: Elliptic Curve defined by y^2 = x^3 + (15*z6^5+5*z6^4+8*z6^3+12*z6^2+11*z6+7)
            over Finite Field in z6 of size 17^6
      To:   Elliptic Curve defined by y^2 = x^3 + z6
            over Finite Field in z6 of size 17^6
    sage: pi6 = EllipticCurveHom_frobenius(E,6)
    sage: pi6hat = pi6.dual(); pi6hat
    Composite morphism of degree 24137569 = 24137569*1:
      From: Elliptic Curve defined by y^2 = x^3 + z6
            over Finite Field in z6 of size 17^6
      To:   Elliptic Curve defined by y^2 = x^3 + z6
            over Finite Field in z6 of size 17^6
    sage: pi6hat.factors()
    (Frobenius endomorphism of degree 24137569 = 17^6:
       From: Elliptic Curve defined by y^2 = x^3 + z6
             over Finite Field in z6 of size 17^6
       To:   Elliptic Curve defined by y^2 = x^3 + z6
             over Finite Field in z6 of size 17^6,
     Elliptic-curve endomorphism of
      Elliptic Curve defined by y^2 = x^3 + z6 over Finite Field in z6 of size 17^6
       Via:  (u,r,s,t) = (2*z6^5 + 10*z6^3 + z6^2 + 8, 0, 0, 0))


TESTS::

    sage: z5, = GF(17^5).gens()
    sage: E = EllipticCurve([z5,1])
    sage: fs = [EllipticCurveHom_frobenius(E)]
    sage: while fs[-1].codomain() != E:
    ....:     fs.append(EllipticCurveHom_frobenius(fs[-1].codomain()))
    sage: fs
    [Frobenius isogeny of degree 17:
       From: Elliptic Curve defined by y^2 = x^3 + z5*x + 1 over Finite Field in z5 of size 17^5
       To:   Elliptic Curve defined by y^2 = x^3 + (9*z5^4+7*z5^3+10*z5^2+z5+14)*x + 1 over Finite Field in z5 of size 17^5,
     Frobenius isogeny of degree 17:
       From: Elliptic Curve defined by y^2 = x^3 + (9*z5^4+7*z5^3+10*z5^2+z5+14)*x + 1 over Finite Field in z5 of size 17^5
       To:   Elliptic Curve defined by y^2 = x^3 + (14*z5^4+7*z5^3+16*z5^2+14*z5+1)*x + 1 over Finite Field in z5 of size 17^5,
     Frobenius isogeny of degree 17:
       From: Elliptic Curve defined by y^2 = x^3 + (14*z5^4+7*z5^3+16*z5^2+14*z5+1)*x + 1 over Finite Field in z5 of size 17^5
       To:   Elliptic Curve defined by y^2 = x^3 + (16*z5^4+6*z5^3+7*z5^2+14*z5+6)*x + 1 over Finite Field in z5 of size 17^5,
     Frobenius isogeny of degree 17:
       From: Elliptic Curve defined by y^2 = x^3 + (16*z5^4+6*z5^3+7*z5^2+14*z5+6)*x + 1 over Finite Field in z5 of size 17^5
       To:   Elliptic Curve defined by y^2 = x^3 + (12*z5^4+14*z5^3+z5^2+4*z5+13)*x + 1 over Finite Field in z5 of size 17^5,
     Frobenius isogeny of degree 17:
       From: Elliptic Curve defined by y^2 = x^3 + (12*z5^4+14*z5^3+z5^2+4*z5+13)*x + 1 over Finite Field in z5 of size 17^5
       To:   Elliptic Curve defined by y^2 = x^3 + z5*x + 1 over Finite Field in z5 of size 17^5]
    sage: prod(fs[::-1])
    Composite morphism of degree 1419857 = 17^5:
      From: Elliptic Curve defined by y^2 = x^3 + z5*x + 1 over Finite Field in z5 of size 17^5
      To:   Elliptic Curve defined by y^2 = x^3 + z5*x + 1 over Finite Field in z5 of size 17^5

::

    sage: EllipticCurveHom_frobenius(EllipticCurve(GF(5),[1,1]), -1)
    Traceback (most recent call last):
    ...
    ValueError: negative powers of Frobenius are not isogenies

::

    sage: EllipticCurveHom_frobenius(EllipticCurve('11a1'))
    Traceback (most recent call last):
    ...
    ValueError: Frobenius isogenies do not exist in characteristic zero

AUTHORS:

- Lorenz Panny (2021): implement :class:`EllipticCurveHom_frobenius`
- Mickaël Montessinos (2021): computing the dual of a Frobenius isogeny
"""

from sage.misc.cachefunc import cached_method
from sage.structure.sequence import Sequence

from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

from sage.rings.finite_rings.finite_field_base import FiniteField

from sage.schemes.elliptic_curves.ell_generic import EllipticCurve_generic
from sage.schemes.elliptic_curves.constructor import EllipticCurve

from sage.schemes.elliptic_curves.hom import EllipticCurveHom, find_post_isomorphism
from sage.schemes.elliptic_curves.ell_curve_isogeny import EllipticCurveIsogeny
from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
from sage.schemes.elliptic_curves.hom_scalar import EllipticCurveHom_scalar


class EllipticCurveHom_frobenius(EllipticCurveHom):

    _degree = None

    def __init__(self, E, power=1):
        r"""
        Construct a Frobenius isogeny on a given curve with a given
        power of the base-ring characteristic.

        Writing `n` for the parameter ``power`` (default: `1`), the
        isogeny is defined by `(x,y) \to (x^{p^n}, y^{p^n})` where
        `p` is the characteristic of the base ring.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(j=GF(11^2).gen())
            sage: EllipticCurveHom_frobenius(E)
            Frobenius isogeny of degree 11:
              From: Elliptic Curve defined by y^2 = x^3 + (2*z2+6)*x + (8*z2+8) over Finite Field in z2 of size 11^2
              To:   Elliptic Curve defined by y^2 = x^3 + (9*z2+3)*x + (3*z2+7) over Finite Field in z2 of size 11^2
            sage: EllipticCurveHom_frobenius(E, 2)
            Frobenius endomorphism of degree 121 = 11^2:
              From: Elliptic Curve defined by y^2 = x^3 + (2*z2+6)*x + (8*z2+8) over Finite Field in z2 of size 11^2
              To:   Elliptic Curve defined by y^2 = x^3 + (2*z2+6)*x + (8*z2+8) over Finite Field in z2 of size 11^2

        TESTS::

            sage: EllipticCurveHom_frobenius(EllipticCurve('11a1'))
            Traceback (most recent call last):
            ...
            ValueError: Frobenius isogenies do not exist in characteristic zero

        ::

            sage: EllipticCurveHom_frobenius(E, -1)
            Traceback (most recent call last):
            ...
            ValueError: negative powers of Frobenius are not isogenies
        """
        if not isinstance(E, EllipticCurve_generic):
            raise ValueError(f'not an elliptic curve: {E}')

        self._base_ring = E.base_ring()

        self._p = self._base_ring.characteristic()
        if self._p == 0:
            raise ValueError('Frobenius isogenies do not exist in characteristic zero')

        self._n = ZZ(power)
        if self._n < 0:
            raise ValueError('negative powers of Frobenius are not isogenies')

        self._degree = self._p ** self._n

        self._domain = E
        as_ = [a**self._degree for a in self._domain.a_invariants()]
        self._codomain = EllipticCurve(as_)

        EllipticCurveHom.__init__(self, self._domain, self._codomain)

        # over finite fields, isogenous curves have the same number of points
        # (see #32786)
        if isinstance(self._base_ring, FiniteField):
            self._domain._fetch_cached_order(self._codomain)
            self._codomain._fetch_cached_order(self._domain)

        self._poly_ring = PolynomialRing(self._base_ring, ['x'], sparse=True)
        self._mpoly_ring = PolynomialRing(self._base_ring, ['x','y'], sparse=True)
        self._xfield = self._poly_ring.fraction_field()
        self._xyfield = self._mpoly_ring.fraction_field()

    def _call_(self, P):
        """
        Evaluate this Frobenius isogeny at a point `P`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: z2 = GF(11^2).gen()
            sage: E = EllipticCurve(j=z2)
            sage: pi = EllipticCurveHom_frobenius(E)
            sage: P = E(7, 9*z2+4)
            sage: pi(P)     # implicit doctest
            (7 : 2*z2 + 7 : 1)
        """
        return self._codomain(*(c**self._degree for c in P))

    def _eval(self, P):
        """
        Less strict evaluation method for internal use.

        In particular, this can be used to evaluate ``self`` at a
        point defined over an extension field.

        INPUT: a sequence of 3 coordinates defining a point on ``self``

        OUTPUT: the result of evaluating ``self`` at the given point

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(11), [1,1])
            sage: pi = EllipticCurveHom_frobenius(E)
            sage: Ebar = E.change_ring(GF(11^6))
            sage: z6 = GF(11^6).gen()
            sage: P = Ebar.lift_x(GF(11^3).gen())
            sage: p = Ebar(6*z6^5 + 8*z6^4 + 8*z6^3 + 6*z6^2 + 10*z6 + 5, 2*z6^5 + 2*z6^4 + 2*z6^3 + 4*z6 + 6, 1)
            sage: Q = pi._eval(P)
            sage: q = Ebar(z6^5 + 3*z6^4 + 3*z6^3 + 6*z6^2 + 9, z6^5 + 10*z6^4 + 10*z6^3 + 5*z6^2 + 4*z6 + 8, 1)
            sage: (P == p and Q == q) or (P == -p and Q == -q)
            True
        """
        if self._domain.defining_polynomial()(*P):
            raise ValueError(f'{P} not on {self._domain}')
        k = Sequence(P).universe()
        return self._codomain.base_extend(k)(*(c**self._degree for c in P))

    def _repr_(self):
        """
        Return basic facts about this Frobenius isogeny as a string.

        TESTS::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: z2 = GF(11^2).gen()
            sage: E = EllipticCurve(j=z2)
            sage: EllipticCurveHom_frobenius(E)
            Frobenius isogeny of degree 11:
              From: Elliptic Curve defined by y^2 = x^3 + (2*z2+6)*x + (8*z2+8) over Finite Field in z2 of size 11^2
              To:   Elliptic Curve defined by y^2 = x^3 + (9*z2+3)*x + (3*z2+7) over Finite Field in z2 of size 11^2
            sage: EllipticCurveHom_frobenius(E, E.base_field().degree())
            Frobenius endomorphism of degree 121 = 11^2:
              From: Elliptic Curve defined by y^2 = x^3 + (2*z2+6)*x + (8*z2+8) over Finite Field in z2 of size 11^2
              To:   Elliptic Curve defined by y^2 = x^3 + (2*z2+6)*x + (8*z2+8) over Finite Field in z2 of size 11^2
        """
        kind = 'endomorphism' if self._codomain == self._domain else 'isogeny'
        degs_str = '' if self._n == 1 else f' = {self._p}^{self._n}'
        return f'Frobenius {kind} of degree {self._degree}{degs_str}:' \
                f'\n  From: {self._domain}' \
                f'\n  To:   {self._codomain}'

    # EllipticCurveHom methods

    def rational_maps(self):
        """
        Return the explicit rational maps defining this Frobenius
        isogeny as (sparse) bivariate rational maps in `x` and `y`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(11), [1,1])
            sage: pi = EllipticCurveHom_frobenius(E, 4)
            sage: pi.rational_maps()
            (x^14641, y^14641)

        TESTS:

        See :issue:`34811`::

            sage: pi.rational_maps()[0].parent()
            Fraction Field of Multivariate Polynomial Ring in x, y over Finite Field of size 11
            sage: pi.rational_maps()[1].parent()
            Fraction Field of Multivariate Polynomial Ring in x, y over Finite Field of size 11
        """
        x,y = self._xyfield.gens()
        return (x**self._degree, y**self._degree)

    def x_rational_map(self):
        """
        Return the `x`-coordinate rational map of this Frobenius
        isogeny as a (sparse) univariate rational map in `x`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(11), [1,1])
            sage: pi = EllipticCurveHom_frobenius(E, 4)
            sage: pi.x_rational_map()
            x^14641

        TESTS:

        See :issue:`34811`::

            sage: pi.x_rational_map().parent()
            Fraction Field of Sparse Univariate Polynomial Ring in x over Finite Field of size 11
        """
        x, = self._xfield.gens()
        return x**self._degree

    def scaling_factor(self):
        r"""
        Return the Weierstrass scaling factor associated to this
        Frobenius morphism.

        The scaling factor is the constant `u` (in the base field)
        such that `\varphi^* \omega_2 = u \omega_1`, where
        `\varphi: E_1\to E_2` is this morphism and `\omega_i` are
        the standard Weierstrass differentials on `E_i` defined by
        `\mathrm dx/(2y+a_1x+a_3)`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(11), [1,1])
            sage: pi = EllipticCurveHom_frobenius(E)
            sage: pi.formal()
            t^11 + O(t^33)
            sage: pi.scaling_factor()
            0
            sage: pi = EllipticCurveHom_frobenius(E, 3)
            sage: pi.formal()
            t^1331 + O(t^1353)
            sage: pi.scaling_factor()
            0
            sage: pi = EllipticCurveHom_frobenius(E, 0)
            sage: pi == E.scalar_multiplication(1)
            True
            sage: pi.scaling_factor()
            1

        The scaling factor lives in the base ring::

            sage: pi.scaling_factor().parent()
            Finite Field of size 11

        ALGORITHM: Inseparable isogenies of degree `>1` have scaling
        factor `0`.
        """
        if self._degree == 1:
            return self._base_ring.one()
        return self._base_ring.zero()

    def kernel_polynomial(self):
        """
        Return the kernel polynomial of this Frobenius isogeny
        as a polynomial in `x`. This method always returns `1`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(11), [1,1])
            sage: pi = EllipticCurveHom_frobenius(E, 5)
            sage: pi.kernel_polynomial()
            1
        """
        return self._poly_ring(1)

    @cached_method
    def dual(self):
        """
        Compute the dual of this Frobenius isogeny.

        This method returns an :class:`EllipticCurveHom` object.

        EXAMPLES:

        An ordinary example::

            sage: from sage.schemes.elliptic_curves.hom_scalar import EllipticCurveHom_scalar
            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(31), [0,1])
            sage: f = EllipticCurveHom_frobenius(E)
            sage: f.dual() * f == EllipticCurveHom_scalar(f.domain(), 31)               # needs sage.symbolic
            True
            sage: f * f.dual() == EllipticCurveHom_scalar(f.codomain(), 31)             # needs sage.symbolic
            True

        A supersingular example::

            sage: E = EllipticCurve(GF(31), [1,0])
            sage: f = EllipticCurveHom_frobenius(E)
            sage: f.dual() * f == EllipticCurveHom_scalar(f.domain(), 31)               # needs sage.symbolic
            True
            sage: f * f.dual() == EllipticCurveHom_scalar(f.codomain(), 31)             # needs sage.symbolic
            True

        TESTS:

        Some random testing (including small characteristic)::

            sage: p = random_prime(50)
            sage: q = p**randrange(1,10)
            sage: n = randrange(20)
            sage: while True:
            ....:     try:
            ....:         E = EllipticCurve([GF(q).random_element() for _ in range(5)])
            ....:         break
            ....:     except ArithmeticError:
            ....:         pass
            sage: f = EllipticCurveHom_frobenius(E, n)
            sage: f.dual() * f == EllipticCurveHom_scalar(E, p**n)                      # needs sage.symbolic
            True
            sage: f * f.dual() == EllipticCurveHom_scalar(f.codomain(), p**n)           # needs sage.symbolic
            True
            sage: f.dual().dual() == f  # known bug -- broken in characteristic 2,3     # needs sage.symbolic
            True
            sage: p in (2,3) or f.dual().dual() == f                                    # needs sage.symbolic
            True

        ALGORITHM:

        - For supersingular curves, the dual of Frobenius is again purely
          inseparable, so we start out with a Frobenius isogeny of equal
          degree in the opposite direction.

        - For ordinary curves, we immediately reduce to the case of prime
          degree. The kernel of the dual is the unique subgroup of size `p`,
          which we compute from the `p`-division polynomial.

        In both cases, we then search for the correct post-isomorphism
        using :meth:`find_post_isomorphism`.
        """
        if self._degree == 1:
            return self

        if self._domain.is_supersingular():
            Phi = EllipticCurveHom_frobenius(self._codomain, self._n)

        else:
            E = self._domain
            poly = self._domain.division_polynomial(self._p)
            ker = self._poly_ring(list(poly)[::self._p]).monic()
            Phis = []
            for _ in range(self._n):
                Ep = EllipticCurve([a**self._p for a in E.a_invariants()])
                Phis.append(EllipticCurveIsogeny(Ep, ker, codomain=E))
                E, ker = Ep, ker.map_coefficients(lambda c: c**self._p)
            Phi = EllipticCurveHom_composite.from_factors(Phis[::-1], self._codomain)

        scalar_mul = EllipticCurveHom_scalar(self._domain, self._degree)
        iso = find_post_isomorphism(Phi * self, scalar_mul)
        return iso * Phi

    def inseparable_degree(self):
        """
        Return the inseparable degree of this Frobenius isogeny.

        Since this class implements only purely inseparable isogenies,
        the inseparable degree equals the degree.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
            sage: E = EllipticCurve(GF(11), [1,1])
            sage: pi = EllipticCurveHom_frobenius(E, 4)
            sage: pi.inseparable_degree()
            14641
            sage: pi.inseparable_degree() == pi.degree()
            True
        """
        return self._degree



# sage_setup: distribution = sagemath-schemes
r"""
Congruence subgroup `\Gamma(N)`
"""

# ****************************************************************************
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import gcd
from sage.groups.matrix_gps.finitely_generated import MatrixGroup
from sage.matrix.constructor import matrix
from sage.misc.misc_c import prod
from sage.modular.arithgroup.congroup_generic import CongruenceSubgroup
from sage.modular.arithgroup.congroup_sl2z import SL2Z
from sage.modular.cusps import Cusp
from sage.rings.finite_rings.integer_mod_ring import Zmod
from sage.rings.integer import GCD_list
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.richcmp import richcmp, richcmp_method

_gamma_cache = {}


def Gamma_constructor(N):
    r"""
    Return the congruence subgroup `\Gamma(N)`.

    EXAMPLES::

        sage: Gamma(5) # indirect doctest
        Congruence Subgroup Gamma(5)
        sage: G = Gamma(23)
        sage: G is Gamma(23)
        True
        sage: TestSuite(G).run()

    Test global uniqueness::

        sage: G = Gamma(17)
        sage: G is loads(dumps(G))
        True
        sage: G2 = sage.modular.arithgroup.congroup_gamma.Gamma_class(17)
        sage: G == G2
        True
        sage: G is G2
        False
    """
    if N == 1:
        return SL2Z
    try:
        return _gamma_cache[N]
    except KeyError:
        _gamma_cache[N] = Gamma_class(N)
        return _gamma_cache[N]


@richcmp_method
class Gamma_class(CongruenceSubgroup):
    r"""
    The principal congruence subgroup `\Gamma(N)`.
    """
    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: Gamma(133)._repr_()
            'Congruence Subgroup Gamma(133)'
        """
        return "Congruence Subgroup Gamma(%s)" % self.level()

    def _latex_(self):
        r"""
        Return the \LaTeX representation of ``self``.

        EXAMPLES::

            sage: Gamma(20)._latex_()
            '\\Gamma(20)'
            sage: latex(Gamma(20))
            \Gamma(20)
        """
        return "\\Gamma(%s)" % self.level()

    def __reduce__(self):
        """
        Used for pickling ``self``.

        EXAMPLES::

            sage: Gamma(5).__reduce__()
            (<function Gamma_constructor at ...>, (5,))
        """
        return Gamma_constructor, (self.level(),)

    def __richcmp__(self, other, op):
        r"""
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: Gamma(3) == SymmetricGroup(8)                                         # needs sage.groups
            False
            sage: Gamma(3) == Gamma1(3)
            False
            sage: Gamma(5) < Gamma(6)
            True
            sage: Gamma(5) == Gamma(5)
            True
            sage: Gamma(3) == Gamma(3).as_permutation_group()                           # needs sage.groups
            True
        """
        if isinstance(other, Gamma_class):
            return richcmp(self.level(), other.level(), op)
        else:
            return NotImplemented

    def index(self):
        r"""
        Return the index of ``self`` in the full modular group. This is given by

        .. MATH::

          \prod_{\substack{p \mid N \\ \text{$p$ prime}}}\left(p^{3e}-p^{3e-2}\right).

        EXAMPLES::

            sage: [Gamma(n).index() for n in [1..19]]
            [1, 6, 24, 48, 120, 144, 336, 384, 648, 720, 1320, 1152, 2184, 2016, 2880, 3072, 4896, 3888, 6840]
            sage: Gamma(32041).index()
            32893086819240
        """
        return prod([p**(3*e-2)*(p*p-1) for (p,e) in self.level().factor()])

    def _contains_sl2(self, a, b, c, d):
        r"""
        EXAMPLES::

            sage: G = Gamma(5)
            sage: [1, 0, -10, 1] in G
            True
            sage: 1 in G
            True
            sage: SL2Z([26, 5, 5, 1]) in G
            True
            sage: SL2Z([1, 1, 6, 7]) in G
            False
        """
        N = self.level()
        # don't need to check d == 1 as this is automatic from det
        return ((a % N == 1) and (b % N == 0) and (c % N == 0))

    def ncusps(self):
        r"""
        Return the number of cusps of this subgroup `\Gamma(N)`.

        EXAMPLES::

            sage: [Gamma(n).ncusps() for n in [1..19]]
            [1, 3, 4, 6, 12, 12, 24, 24, 36, 36, 60, 48, 84, 72, 96, 96, 144, 108, 180]
            sage: Gamma(30030).ncusps()
            278691840
            sage: Gamma(2^30).ncusps()
            432345564227567616
        """
        n = self.level()
        if n == 1:
            return ZZ(1)
        if n == 2:
            return ZZ(3)
        return prod([p**(2*e) - p**(2*e-2) for (p,e) in n.factor()])//2

    def nirregcusps(self):
        r"""
        Return the number of irregular cusps of ``self``. For principal
        congruence subgroups this is always 0.

        EXAMPLES::

            sage: Gamma(17).nirregcusps()
            0
        """
        return 0

    def _find_cusps(self):
        r"""
        Calculate the reduced representatives of the equivalence classes of
        cusps for this group. Adapted from code by Ron Evans.

        EXAMPLES::

            sage: Gamma(8).cusps() # indirect doctest
            [0, 1/4, 1/3, 3/8, 1/2, 2/3, 3/4, 1, 4/3, 3/2, 5/3, 2, 7/3, 5/2, 8/3, 3, 7/2, 11/3, 4, 14/3, 5, 6, 7, Infinity]
        """
        n = self.level()
        C = [QQ(x) for x in range(n)]

        n0 = n//2
        n1 = (n+1)//2

        for r in range(1, n1):
            if r > 1 and gcd(r,n) == 1:
                C.append(ZZ(r)/ZZ(n))
            if n0 == n/2 and gcd(r,n0) == 1:
                C.append(ZZ(r)/ZZ(n0))

        for s in range(2,n1):
            for r in range(1, 1+n):
                if GCD_list([s,r,n]) == 1:
                    # GCD_list is ~40x faster than gcd, since gcd wastes loads
                    # of time initialising a Sequence type.
                    u,v = _lift_pair(r,s,n)
                    C.append(ZZ(u)/ZZ(v))

        return [Cusp(x) for x in sorted(C)] + [Cusp(1,0)]

    def reduce_cusp(self, c):
        r"""
        Calculate the unique reduced representative of the equivalence of the
        cusp `c` modulo this group. The reduced representative of an
        equivalence class is the unique cusp in the class of the form `u/v`
        with `u, v \ge 0` coprime, `v` minimal, and `u` minimal for that `v`.

        EXAMPLES::

            sage: Gamma(5).reduce_cusp(1/5)
            Infinity
            sage: Gamma(5).reduce_cusp(7/8)
            3/2
            sage: Gamma(6).reduce_cusp(4/3)
            2/3

        TESTS::

            sage: G = Gamma(50)
            sage: all(c == G.reduce_cusp(c) for c in G.cusps())
            True

        We test that :issue:`36163` is fixed::

            sage: Gamma(7).reduce_cusp(Cusp(6,7))
            Infinity
        """
        N = self.level()
        c = Cusp(c)
        u,v = c.numerator() % N, c.denominator() % N
        if (v > N//2) or (2*v == N and u > N//2):
            u,v = -u,-v
        u,v = _lift_pair(u,v,N)
        return Cusp(u,v)

    def are_equivalent(self, x, y, trans=False):
        r"""
        Check if the cusps `x` and `y` are equivalent under the action of this group.

        ALGORITHM: The cusps `u_1 / v_1` and `u_2 / v_2` are equivalent modulo
        `\Gamma(N)` if and only if `(u_1, v_1) = \pm (u_2, v_2) \bmod N`.

        EXAMPLES::

            sage: Gamma(7).are_equivalent(Cusp(2/3), Cusp(5/4))
            True
        """
        if trans:
            return CongruenceSubgroup.are_equivalent(self, x,y,trans=trans)
        N = self.level()
        u1,v1 = (x.numerator() % N, x.denominator() % N)
        u2,v2 = (y.numerator(), y.denominator())

        return ((u1,v1) == (u2 % N, v2 % N)) or ((u1,v1) == (-u2 % N, -v2 % N))

    def nu3(self):
        r"""
        Return the number of elliptic points of order 3 for this arithmetic
        subgroup. Since this subgroup is `\Gamma(N)` for `N \ge 2`, there are
        no such points, so we return 0.

        EXAMPLES::

            sage: Gamma(89).nu3()
            0
        """
        return 0

    # We don't need to override nu2, since the default nu2 implementation knows
    # that nu2 = 0 for odd subgroups.

    def image_mod_n(self):
        r"""
        Return the image of this group modulo `N`, as a subgroup of `SL(2, \ZZ
        / N\ZZ)`. This is just the trivial subgroup.

        EXAMPLES::

            sage: Gamma(3).image_mod_n()
            Matrix group over Ring of integers modulo 3 with 1 generators (
            [1 0]
            [0 1]
            )
        """
        return MatrixGroup([matrix(Zmod(self.level()), 2, 2, 1)])


def is_Gamma(x):
    r"""
    Return ``True`` if x is a congruence subgroup of type Gamma.

    EXAMPLES::

        sage: from sage.modular.arithgroup.congroup_gamma import Gamma_class
        sage: isinstance(Gamma0(13), Gamma_class)
        False
        sage: isinstance(Gamma(4), Gamma_class)
        True
    """
    from sage.misc.superseded import deprecation
    deprecation(38035, "The function is_Gamma is deprecated; use 'isinstance(..., Gamma_class)' instead.")
    return isinstance(x, Gamma_class)


def _lift_pair(U, V, N):
    r"""
    Utility function. Given integers ``U, V, N``, with `N \ge 1` and `{\rm
    gcd}(U, V, N) = 1`, return a pair `(u, v)` congruent to `(U, V) \bmod N`,
    such that `{\rm gcd}(u,v) = 1`, `u, v \ge 0`, `v` is as small as possible,
    and `u` is as small as possible for that `v`.

    *Warning*: As this function is for internal use, it does not do a
    preliminary sanity check on its input, for efficiency. It will recover
    reasonably gracefully if ``(U, V, N)`` are not coprime, but only after
    wasting quite a lot of cycles!

    EXAMPLES::

        sage: from sage.modular.arithgroup.congroup_gamma import _lift_pair
        sage: _lift_pair(2,4,7)
        (9, 4)
        sage: _lift_pair(2,4,8) # don't do this
        Traceback (most recent call last):
        ...
        ValueError: (U, V, N) must be coprime
    """
    u = U % N
    v = V % N
    if v == 0:
        if u == 1 or u == N-1:
            return (1,0)
        else:
            v = N
    while gcd(u, v) > 1:
        u = u+N
        if u > N*v:
            raise ValueError("(U, V, N) must be coprime")
    return (u, v)



# sage_setup: distribution = sagemath-gsl
# sage.doctest: known bug: windows (crash - https://github.com/passagemath/passagemath/issues/1892#issue-3726343166)
r"""
Real Interpolation using GSL
"""

from cysignals.memory cimport sig_malloc, sig_free
from cysignals.signals cimport sig_on, sig_off


cdef class Spline:
    """
    Create a spline interpolation object.

    Given a list `v` of pairs, ``s = spline(v)`` is an object ``s`` such that
    `s(x)` is the value of the spline interpolation through the points
    in `v` at the point `x`.

    The values in `v` do not have to be sorted.  Moreover, one can append
    values to `v`, delete values from `v`, or change values in `v`, and the
    spline is recomputed.

    EXAMPLES::

        sage: S = spline([(0, 1), (1, 2), (4, 5), (5, 3)]); S
        [(0, 1), (1, 2), (4, 5), (5, 3)]
        sage: S(1.5)
        2.76136363636...

    Changing the points of the spline causes the spline to be recomputed::

        sage: S[0] = (0, 2); S
        [(0, 2), (1, 2), (4, 5), (5, 3)]
        sage: S(1.5)
        2.507575757575...

    We may delete interpolation points of the spline::

        sage: del S[2]; S
        [(0, 2), (1, 2), (5, 3)]
        sage: S(1.5)
        2.04296875

    We may append to the list of interpolation points::

        sage: S.append((4, 5)); S
        [(0, 2), (1, 2), (5, 3), (4, 5)]
        sage: S(1.5)
        2.507575757575...

    If we set the `n`-th interpolation point, where `n` is larger than
    ``len(S)``, then points `(0, 0)` will be inserted between the
    interpolation points and the point to be added::

        sage: S[6] = (6, 3); S
        [(0, 2), (1, 2), (5, 3), (4, 5), (0, 0), (0, 0), (6, 3)]

    This example is in the GSL documentation::

        sage: v = [(i + RDF(i).sin()/2, i + RDF(i^2).cos()) for i in range(10)]
        sage: s = spline(v)
        sage: show(point(v) + plot(s,0,9, hue=.8))                                      # needs sage.plot

    We compute the area underneath the spline::

        sage: s.definite_integral(0, 9)
        41.196516041067...

    The definite integral is additive::

        sage: s.definite_integral(0, 4) + s.definite_integral(4, 9)
        41.196516041067...

    Switching the order of the bounds changes the sign of the integral::

        sage: s.definite_integral(9, 0)
        -41.196516041067...

    We compute the first and second-order derivatives at a few points::

        sage: s.derivative(5)
        -0.1623008526180...
        sage: s.derivative(6)
        0.2099798628571...
        sage: s.derivative(5, order=2)
        -3.0874707456138...
        sage: s.derivative(6, order=2)
        2.6187684827485...

    Only the first two derivatives are supported::

        sage: s.derivative(4, order=3)
        Traceback (most recent call last):
        ...
        ValueError: Order of derivative must be 1 or 2.
    """
    def __init__(self, v=None):
        """
        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)]); S
            [(1, 1), (2, 3), (4, 5)]
            sage: type(S)
            <class 'sage.calculus.interpolation.Spline'>
        """
        self.v = [] if v is None else list(v)
        self.started = 0

    def __dealloc__(self):
        self.stop_interp()

    def __setitem__(self, int i, xy):
        """
        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)]); S
            [(1, 1), (2, 3), (4, 5)]
            sage: S(1.5)
            2.0625

        Replace `0`-th point, which changes the spline::

            sage: S[0] = (0,1); S
            [(0, 1), (2, 3), (4, 5)]
            sage: S(1.5)
            2.5

        If you set the `n`-th point and `n` is larger than ``len(S)``,
        then `(0,0)` points are inserted and the `n`-th entry is set
        (which may be a weird thing to do, but that is what happens)::

            sage: S[4] = (6,10)
            sage: S
            [(0, 1), (2, 3), (4, 5), (0, 0), (6, 10)]
        """
        cdef int j
        if i < len(self.v):
            self.v[i] = xy
        else:
            for j in range(len(self.v), i + 1):
                self.v.append((0, 0))
            self.v[i] = xy
        self.stop_interp()

    def __getitem__(self, int i):
        """
        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)]); S[0]
            (1, 1)
            sage: S[-1]
            (4, 5)
            sage: S[1]
            (2, 3)
        """
        return self.v[i]

    def __delitem__(self, int i):
        """
        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)]); S
            [(1, 1), (2, 3), (4, 5)]
            sage: del S[1]
            sage: S
            [(1, 1), (4, 5)]

        The spline is recomputed when points are deleted (:issue:`13519`)::

            sage: S = spline([(1,1), (2,3), (4,5), (5, 5)]); S
            [(1, 1), (2, 3), (4, 5), (5, 5)]
            sage: S(3)
            4.375
            sage: del S[0]; S
            [(2, 3), (4, 5), (5, 5)]
            sage: S(3)
            4.25
        """
        del self.v[i]
        self.stop_interp()

    def append(self, xy):
        """
        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)]); S.append((5,7)); S
            [(1, 1), (2, 3), (4, 5), (5, 7)]

        The spline is recomputed when points are appended (:issue:`13519`)::

            sage: S = spline([(1,1), (2,3), (4,5)]); S
            [(1, 1), (2, 3), (4, 5)]
            sage: S(3)
            4.25
            sage: S.append((5, 5)); S
            [(1, 1), (2, 3), (4, 5), (5, 5)]
            sage: S(3)
            4.375
        """
        self.v.append(xy)
        self.stop_interp()

    def list(self):
        """
        Underlying list of points that this spline goes through.

        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)]); S.list()
            [(1, 1), (2, 3), (4, 5)]

        This is a copy of the list, not a reference (:issue:`13530`)::

            sage: S = spline([(1,1), (2,3), (4,5)])
            sage: L = S.list(); L
            [(1, 1), (2, 3), (4, 5)]
            sage: L[2] = (3, 2)
            sage: L
            [(1, 1), (2, 3), (3, 2)]
            sage: S.list()
            [(1, 1), (2, 3), (4, 5)]
        """
        return self.v[:]

    def __len__(self):
        """
        Number of points that the spline goes through.

        EXAMPLES::

            sage: len(spline([(1,1), (2,3), (4,5)]))
            3
        """
        return len(self.v)

    def __repr__(self):
        """
        EXAMPLES::

            sage: spline([(1,1), (2,3), (4,5)]).__repr__()
            '[(1, 1), (2, 3), (4, 5)]'
        """
        return str(self.v)

    cdef start_interp(self):
        if self.started:
            sig_free(self.x)
            sig_free(self.y)
            return
        v = list(self.v)
        v.sort()
        n = len(v)
        if n < 3:
            raise RuntimeError("must have at least 3 points in order to interpolate")
        self.x = <double*> sig_malloc(n*sizeof(double))
        if self.x == <double*>0:
            raise MemoryError
        self.y = <double*> sig_malloc(n*sizeof(double))
        if self.y == <double*>0:
            sig_free(self.x)
            raise MemoryError

        cdef int i
        for i in range(n):
            self.x[i] = v[i][0]
            self.y[i] = v[i][1]

        self.acc = gsl_interp_accel_alloc ()
        self.spline = gsl_spline_alloc (gsl_interp_cspline, n)
        gsl_spline_init (self.spline, self.x, self.y, n)
        self.started = 1

    cdef stop_interp(self):
        if not self.started:
            return
        sig_free(self.x)
        sig_free(self.y)
        gsl_spline_free (self.spline)
        gsl_interp_accel_free (self.acc)
        self.started = 0

    def __call__(self, double x):
        """
        Value of the spline function at `x`.

        EXAMPLES::

            sage: S = spline([(1,1), (2,3), (4,5)])
            sage: S(1)
            1.0
            sage: S(2)
            3.0
            sage: S(4)
            5.0
            sage: S(3.5)
            4.65625
        """
        if not self.started:
            self.start_interp()
        sig_on()
        y = gsl_spline_eval(self.spline, x, self.acc)
        sig_off()
        return y

    def derivative(self, double x, int order=1):
        """
        Value of the first or second derivative of the spline at `x`.

        INPUT:

        - ``x`` -- value at which to evaluate the derivative

        - ``order`` -- (default: 1) order of the derivative; must be 1 or 2

        EXAMPLES:

        We draw a cubic spline through three points and compute the
        derivatives::

            sage: s = spline([(0, 0), (2, 3), (4, 0)])
            sage: s.derivative(0)
            2.25
            sage: s.derivative(2)
            0.0
            sage: s.derivative(4)
            -2.25
            sage: s.derivative(1, order=2)
            -1.125
            sage: s.derivative(3, order=2)
            -1.125
        """
        if (order!=1) and (order!=2):
            raise ValueError("Order of derivative must be 1 or 2.")

        if not self.started:
            self.start_interp()
        sig_on()

        if order == 1:
            d = gsl_spline_eval_deriv(self.spline, x, self.acc)
        else:
            d = gsl_spline_eval_deriv2(self.spline, x, self.acc)

        sig_off()
        return d

    def definite_integral(self, double a, double b):
        """
        Value of the definite integral between `a` and `b`.

        INPUT:

        - ``a`` -- lower bound for the integral

        - ``b`` -- upper bound for the integral

        EXAMPLES:

        We draw a cubic spline through three points and compute the
        area underneath the curve::

            sage: s = spline([(0, 0), (1, 3), (2, 0)])
            sage: s.definite_integral(0, 2)
            3.75
            sage: s.definite_integral(0, 1)
            1.875
            sage: s.definite_integral(0, 1) + s.definite_integral(1, 2)
            3.75
            sage: s.definite_integral(2, 0)
            -3.75
        """
        # GSL chokes when the upper bound is smaller than the lower bound
        bounds_swapped = False
        if b < a:
            a, b = b, a
            bounds_swapped = True

        if not self.started:
            self.start_interp()
        sig_on()
        I = gsl_spline_eval_integ(self.spline, a, b, self.acc)
        sig_off()

        if bounds_swapped:
            I = -I
        return I

spline = Spline



# sage_setup: distribution = sagemath-modules
"""
Miscellaneous matrix functions
"""

# ****************************************************************************
#       Copyright (C) 2008 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.fields import Fields
_Fields = Fields()


def row_iterator(A):
    for i in range(A.nrows()):
        yield A.row(i)


def prm_mul(p1, p2, mask_free, prec):
    """
    Return the product of ``p1`` and ``p2``, putting free variables in
    ``mask_free`` to `1`.

    This function is mainly use as a subroutine of
    :func:`permanental_minor_polynomial`.

    INPUT:

    - `p1,p2` -- polynomials as dictionaries

    - ``mask_free`` -- integer mask that give the list of free variables
      (the `i`-th variable is free if the `i`-th bit of ``mask_free`` is `1`)

    - ``prec`` -- if ``prec`` is not ``None``, truncate the product at precision ``prec``

    EXAMPLES::

        sage: from sage.matrix.matrix_misc import prm_mul
        sage: t = polygen(ZZ, 't')
        sage: p1 = {0: 1, 1: t, 4: t}
        sage: p2 = {0: 1, 1: t, 2: t}
        sage: prm_mul(p1, p2, 1, None)
        {0: 2*t + 1, 2: t^2 + t, 4: t^2 + t, 6: t^2}
    """
    p = {}
    if not p2:
        return p
    for exp1, v1 in p1.items():
        if v1.is_zero():
            continue
        for exp2, v2 in p2.items():
            if exp1 & exp2:
                continue
            v = v1 * v2
            if prec is not None:
                v._unsafe_mutate(prec, 0)
            exp = exp1 | exp2
            exp = exp ^ (exp & mask_free)
            if exp not in p:
                p[exp] = v
            else:
                p[exp] += v
    return p


def permanental_minor_polynomial(A, permanent_only=False, var='t', prec=None):
    r"""
    Return the polynomial of the sums of permanental minors of ``A``.

    INPUT:

    - ``A`` -- a matrix

    - ``permanent_only`` -- if ``True``, return only the permanent of `A`

    - ``var`` -- name of the polynomial variable

    - ``prec`` -- if prec is not None, truncate the polynomial at precision `prec`


    The polynomial of the sums of permanental minors is

    .. MATH::

        \sum_{i=0}^{min(nrows, ncols)} p_i(A) x^i

    where `p_i(A)` is the `i`-th permanental minor of `A` (that can also be
    obtained through the method
    :meth:`~sage.matrix.matrix2.Matrix.permanental_minor` via
    ``A.permanental_minor(i)``).

    The algorithm implemented by that function has been developed by P. Butera
    and M. Pernici, see [BP2015]_. Its complexity is `O(2^n m^2 n)` where `m` and
    `n` are the number of rows and columns of `A`.  Moreover, if `A` is a banded
    matrix with width `w`, that is `A_{ij}=0` for `|i - j| > w` and `w < n/2`,
    then the complexity of the algorithm is `O(4^w (w+1) n^2)`.

    INPUT:

    - ``A`` -- matrix

    - ``permanent_only`` -- boolean (default: ``False``); if ``True``, only the
      permanent is computed (might be faster)

    - ``var`` -- a variable name

    EXAMPLES::

        sage: from sage.matrix.matrix_misc import permanental_minor_polynomial
        sage: m = matrix([[1,1],[1,2]])
        sage: permanental_minor_polynomial(m)
        3*t^2 + 5*t + 1
        sage: permanental_minor_polynomial(m, permanent_only=True)
        3
        sage: permanental_minor_polynomial(m, prec=2)
        5*t + 1

    ::

        sage: M = MatrixSpace(ZZ,4,4)
        sage: A = M([1,0,1,0,1,0,1,0,1,0,10,10,1,0,1,1])
        sage: permanental_minor_polynomial(A)
        84*t^3 + 114*t^2 + 28*t + 1
        sage: [A.permanental_minor(i) for i in range(5)]
        [1, 28, 114, 84, 0]

    An example over `\QQ`::

        sage: M = MatrixSpace(QQ,2,2)
        sage: A = M([1/5,2/7,3/2,4/5])
        sage: permanental_minor_polynomial(A, True)
        103/175

    An example with polynomial coefficients::

        sage: R.<a> = PolynomialRing(ZZ)
        sage: A = MatrixSpace(R,2)([[a,1], [a,a+1]])
        sage: permanental_minor_polynomial(A, True)
        a^2 + 2*a

    A usage of the ``var`` argument::

        sage: m = matrix(ZZ,4,[0,1,2,3,1,2,3,0,2,3,0,1,3,0,1,2])
        sage: permanental_minor_polynomial(m, var='x')
        164*x^4 + 384*x^3 + 172*x^2 + 24*x + 1

    ALGORITHM:

        The permanent `perm(A)` of a `n \times n` matrix `A` is the coefficient
        of the `x_1 x_2 \ldots x_n` monomial in

        .. MATH::

            \prod_{i=1}^n \left( \sum_{j=1}^n A_{ij} x_j \right)

        Evaluating this product one can neglect `x_i^2`, that is `x_i`
        can be considered to be nilpotent of order `2`.

        To formalize this procedure, consider the algebra
        `R = K[\eta_1, \eta_2, \ldots, \eta_n]` where the `\eta_i` are
        commuting, nilpotent of order `2` (i.e. `\eta_i^2 = 0`).
        Formally it is the quotient ring of the polynomial
        ring in `\eta_1, \eta_2, \ldots, \eta_n` quotiented by the ideal
        generated by the `\eta_i^2`.

        We will mostly consider the ring `R[t]` of polynomials over `R`. We
        denote a generic element of `R[t]` by `p(\eta_1, \ldots, \eta_n)` or
        `p(\eta_{i_1}, \ldots, \eta_{i_k})` if we want to emphasize that some
        monomials in the `\eta_i` are missing.

        Introduce an "integration" operation `\langle p \rangle` over `R` and
        `R[t]` consisting in the sum of the coefficients of the non-vanishing
        monomials in `\eta_i` (i.e. the result of setting all variables `\eta_i`
        to `1`). Let us emphasize that this is *not* a morphism of algebras as
        `\langle \eta_1 \rangle^2 = 1` while `\langle \eta_1^2 \rangle = 0`!

        Let us consider an example of computation.
        Let `p_1 = 1 + t \eta_1 + t \eta_2` and
        `p_2 = 1 + t \eta_1 + t \eta_3`. Then

        .. MATH::

            p_1 p_2 = 1 + 2t \eta_1 +
                    t (\eta_2 + \eta_3) +
                    t^2 (\eta_1 \eta_2 + \eta_1 \eta_3 + \eta_2 \eta_3)

        and

        .. MATH::

            \langle p_1 p_2 \rangle = 1 + 4t + 3t^2

        In this formalism, the permanent is just

        .. MATH::

            perm(A) = \langle \prod_{i=1}^n \sum_{j=1}^n A_{ij} \eta_j \rangle

        A useful property of `\langle . \rangle` which makes this algorithm
        efficient for band matrices is the following: let
        `p_1(\eta_1, \ldots, \eta_n)` and `p_2(\eta_j, \ldots, \eta_n)` be
        polynomials in `R[t]` where `j \ge 1`.  Then one has

        .. MATH::

            \langle p_1(\eta_1, \ldots, \eta_n) p_2 \rangle =
            \langle p_1(1, \ldots, 1, \eta_j, \ldots, \eta_n) p_2 \rangle

        where `\eta_1,..,\eta_{j-1}` are replaced by `1` in `p_1`. Informally,
        we can "integrate" these variables *before* performing the product. More
        generally, if a monomial `\eta_i` is missing in one of the terms of a
        product of two terms, then it can be integrated in the other term.

        Now let us consider an `m \times n` matrix with `m \leq n`. The *sum of
        permanental `k`-minors of `A`* is

        .. MATH::

            perm(A, k) = \sum_{r,c} perm(A_{r,c})

        where the sum is over the `k`-subsets `r` of rows and `k`-subsets `c` of
        columns and `A_{r,c}` is the submatrix obtained from `A` by keeping only
        the rows `r` and columns `c`. Of course
        `perm(A, \min(m,n)) = perm(A)` and note that `perm(A,1)` is just the sum
        of all entries of the matrix.

        The generating function of these sums of permanental minors is

        .. MATH::

            g(t) = \left\langle
            \prod_{i=1}^m \left(1 + t \sum_{j=1}^n A_{ij} \eta_j\right)
            \right\rangle

        In fact the `t^k` coefficient of `g(t)` corresponds to choosing
        `k` rows of `A`;  `\eta_i` is associated to the `i`-th column;
        nilpotency avoids having twice the same column in a product of `A`'s.

        For more details, see the article [BP2015]_.

        From a technical point of view, the product in
        `K[\eta_1, \ldots, \eta_n][t]` is implemented as a subroutine in
        :func:`prm_mul`. The indices of the rows and columns actually start at
        `0`, so the variables are  `\eta_0, \ldots, \eta_{n-1}`. Polynomials are
        represented in dictionary form: to a variable `\eta_i` is associated
        the key `2^i` (or in Python ``1 << i``). The keys associated to products
        are obtained by considering the development in base `2`: to the monomial
        `\eta_{i_1} \ldots \eta_{i_k}` is associated the key
        `2^{i_1} + \ldots + 2^{i_k}`. So the product `\eta_1 \eta_2` corresponds
        to the key `6 = (110)_2` while `\eta_0 \eta_3` has key `9 = (1001)_2`.
        In particular all operations on monomials are implemented via bitwise
        operations on the keys.
    """
    if permanent_only:
        prec = None
    elif prec is not None:
        prec = int(prec)
        if prec == 0:
            raise ValueError('the argument `prec` must be a positive integer')

    from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

    K = PolynomialRing(A.base_ring(), var)
    nrows = A.nrows()
    ncols = A.ncols()
    A = A.rows()
    p = {0: K.one()}
    t = K.gen()
    vars_to_do = list(range(ncols))
    for i in range(nrows):
        # build the polynomial p1 = 1 + t sum A_{ij} eta_j
        if permanent_only:
            p1 = {}
        else:
            p1 = {0: K.one()}
        a = A[i]   # the i-th row of A
        for j in range(len(a)):
            if a[j]:
                p1[1 << j] = a[j] * t

        # make the product with the preceding polynomials, taking care of
        # variables that can be integrated
        mask_free = 0
        j = 0
        while j < len(vars_to_do):
            jj = vars_to_do[j]
            if all(A[k][jj] == 0 for k in range(i+1, nrows)):
                mask_free += 1 << jj
                vars_to_do.remove(jj)
            else:
                j += 1
        p = prm_mul(p, p1, mask_free, prec)

    if not p:
        return K.zero()

    if len(p) != 1 or 0 not in p:
        raise RuntimeError("Something is wrong! Certainly a problem in the"
                           " algorithm... please contact sage-devel@googlegroups.com")

    p = p[0]
    return p[min(nrows, ncols)] if permanent_only else p



# sage_setup: distribution = sagemath-modules
"""
Clifford algebra elements
"""
from sage.modules.with_basis.indexed_element cimport IndexedFreeModuleElement
from sage.data_structures.bitset cimport FrozenBitset

cdef class CliffordAlgebraElement(IndexedFreeModuleElement):
    cdef CliffordAlgebraElement _mul_self_term(self, FrozenBitset supp, coeff)
    cdef CliffordAlgebraElement _mul_term_self(self, FrozenBitset supp, coeff)

cdef class ExteriorAlgebraElement(CliffordAlgebraElement):
    pass

cdef class CohomologyRAAGElement(CliffordAlgebraElement):
    pass



r"""
Generic Convolution

Asymptotically fast convolution of lists over any commutative ring
in which the multiply-by-two map is injective. (More precisely, if
`x \in R`, and `x = 2^k*y` for some `k \geq 0`, we require that
`R(x/2^k)` returns `y`.)

The main function to be exported is :func:`convolution`.

EXAMPLES::

    sage: convolution([1, 2, 3, 4, 5], [6, 7])
    [6, 19, 32, 45, 58, 35]

The convolution function is reasonably fast, even though it is written
in pure Python. For example, the following takes less than a second::

    sage: v = convolution(list(range(1000)), list(range(1000)))

ALGORITHM:

Converts the problem to multiplication in the ring
`S[x]/(x^M - 1)`, where `S = R[y]/(y^K + 1)` (where
`R` is the original base ring). Performs FFT with respect
to the roots of unity `1, y, y^2, \ldots, y^{2K-1}` in
`S`. The FFT/IFFT are accomplished with just additions and
subtractions and rotating python lists. (I think this algorithm is
essentially due to Schonhage, not completely sure.) The pointwise
multiplications are handled recursively, switching to a classical
algorithm at some point.

Complexity is O(n log(n) log(log(n))) additions/subtractions in R
and O(n log(n)) multiplications in R.

AUTHORS:

- David Harvey (2007-07): first implementation

- William Stein: editing the docstrings for inclusion in Sage.
"""
# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#                          David Harvey <dmharvey@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.structure.element import parent
from math import log, ceil

# -------------------------------------------------------------------
#      main exported routine


def convolution(L1, L2):
    """
    Return convolution of non-empty lists L1 and L2.

    L1 and L2 may have arbitrary lengths.

    EXAMPLES::

        sage: convolution([1, 2, 3], [4, 5, 6, 7])
        [4, 13, 28, 34, 32, 21]

    TESTS::

        sage: R = Integers(47)
        sage: L1 = [R.random_element() for _ in range(1000)]
        sage: L2 = [R.random_element() for _ in range(3756)]
        sage: L3 = convolution(L1, L2)
        sage: L3[2000] == sum([L1[i] * L2[2000-i] for i in range(1000)])
        True
        sage: len(L3) == 1000 + 3756 - 1
        True
    """
    if not L1 or not L2:
        raise ValueError("cannot compute convolution of empty lists")
    if len(L1) <= 100 and len(L2) <= 100:   # very arbitrary cutoff
        return _convolution_naive(L1, L2)
    return _convolution_fft(L1, L2)


# -------------------------------------------------------------------
#      naive convolution and negacyclic convolutions


def _convolution_naive(L1, L2):
    """
    Return convolution of non-empty lists L1 and L2, using naive algorithm.

    L1 and L2 may have arbitrary lengths.

    EXAMPLES::

        sage: from sage.rings.polynomial.convolution import _convolution_naive
        sage: _convolution_naive([2], [3])
        [6]
        sage: _convolution_naive([2, 5], [3])
        [6, 15]
        sage: _convolution_naive([2], [3, 6])
        [6, 12]
        sage: _convolution_naive([1, 2, 3], [4, 5, 6, 7])
        [4, 13, 28, 34, 32, 21]
        sage: _convolution_naive([4, 5, 6, 7], [1, 2, 3])
        [4, 13, 28, 34, 32, 21]
    """
    assert len(L1) and len(L2)

    m1 = len(L1)
    m2 = len(L2)

    return [sum([L1[i] * L2[k - i]
            for i in range(max(0, k - m2 + 1), min(k + 1, m1))])
            for k in range(m1 + m2 - 1)]


def _negaconvolution_naive(L1, L2):
    """
    Negacyclic convolution of L1 and L2, using naive algorithm.

    L1 and L2 must be the same length.

    EXAMPLES::

        sage: from sage.rings.polynomial.convolution import _negaconvolution_naive
        sage: from sage.rings.polynomial.convolution import _convolution_naive
        sage: _negaconvolution_naive([2], [3])
        [6]
        sage: _convolution_naive([1, 2, 3], [3, 4, 5])
        [3, 10, 22, 22, 15]
        sage: _negaconvolution_naive([1, 2, 3], [3, 4, 5])
        [-19, -5, 22]
    """
    assert len(L1)
    assert len(L1) == len(L2)

    N = len(L1)
    return [sum([L1[i] * L2[j - i] for i in range(j + 1)]) -
            sum([L1[i] * L2[N + j - i]
                 for i in range(j + 1, N)]) for j in range(N)]


# -------------------------------------------------------------------
#      FFT/IFFT routines


def _forward_butterfly(L1, L2, r):
    r"""
    L1 and L2 are both lists of length K, and
    `0 \leq r \leq K`. They represent polynomials in
    `S = R[y]/(y^K + 1)`. This function returns
    `(L_1 + y^r L_2, L_1 - y^r L_2)`, as a list.
    """
    assert len(L1) == len(L2)
    assert 0 <= r <= len(L1)

    K = len(L1)
    return [L1[i] - L2[i + K - r] for i in range(r)] + \
        [L1[i] + L2[i - r] for i in range(r, K)], \
        [L1[i] + L2[i + K - r] for i in range(r)] + \
        [L1[i] - L2[i - r] for i in range(r, K)]


def _inverse_butterfly(L1, L2, r):
    r"""
    L1 and L2 are both lists of length `K`, and
    `0 \leq r \leq K`. They represent polynomials in
    `S = R[y]/(y^K + 1)`. This function returns
    `(L_1 + L_2, y^{-r}*(L_1 - L_2))`, as a list.
    """
    assert len(L1) == len(L2)
    assert 0 <= r <= len(L1)

    K = len(L1)
    return [L1[i] + L2[i] for i in range(K)], \
        [L1[i] - L2[i] for i in range(r, K)] + \
        [L2[i] - L1[i] for i in range(r)]


def _fft(L, K, start, depth, root):
    r"""
    L is a list of length `M = 2^m`, each entry a list of
    length `K = 2^k`.

    This function only operates on the [start, start + D) portion of L,
    where `D = 2^\text{depth}`. This portion is interpreted as
    a polynomial in `S[x]/(x^D - y^(2*root))`, where
    `S = R[y]/(y^K + 1)`.

    This function performs an inplace FFT, i.e. evaluates the
    polynomial at x = each `D`-th root of unity in S (namely the powers
    of `y^{2K/D}`), with results in bit-reversed order.
    """
    half = 1 << (depth - 1)
    start2 = start + half

    # reduce mod (x^(D/2) - y^root) and mod (x^(D/2) + y^root)
    for i in range(half):
        L[start + i], L[start2 + i] = \
            _forward_butterfly(L[start + i], L[start2 + i], root)

    # recurse into each half
    if depth >= 2:
        _fft(L, K, start, depth - 1, root >> 1)
        _fft(L, K, start2, depth - 1, (root + K) >> 1)


def _ifft(L, K, start, depth, root):
    r"""
    Inverse operation of ``_fft_trunc()`` (except that
    result is a factor of ``2^depth`` too big)
    """
    half = 1 << (depth - 1)
    start2 = start + half

    # recurse into each half
    if depth >= 2:
        _ifft(L, K, start, depth - 1, root >> 1)
        _ifft(L, K, start2, depth - 1, (root + K) >> 1)

    # CRT together (x^(D/2) - y^root) and mod (x^(D/2) + y^root)
    for i in range(half):
        L[start + i], L[start2 + i] = \
            _inverse_butterfly(L[start + i], L[start2 + i], root)

# -------------------------------------------------------------------
#      splitting and recombining routines


def _split(L, m, k):
    """
    Assumes L is a list of length `2^{m+k-1}`. Splits it into
    `2^m` lists of length `2^{k-1}`, returned as a list
    of lists. Each list is zero padded up to length `2^k`.

    EXAMPLES::

        sage: from sage.rings.polynomial.convolution import _split
        sage: _split([1, 2, 3, 4, 5, 6, 7, 8], 2, 2)
        [[1, 2, 0, 0], [3, 4, 0, 0], [5, 6, 0, 0], [7, 8, 0, 0]]
        sage: _split([1, 2, 3, 4, 5, 6, 7, 8], 1, 3)
        [[1, 2, 3, 4, 0, 0, 0, 0], [5, 6, 7, 8, 0, 0, 0, 0]]
        sage: _split([1, 2, 3, 4, 5, 6, 7, 8], 3, 1)
        [[1, 0], [2, 0], [3, 0], [4, 0], [5, 0], [6, 0], [7, 0], [8, 0]]
    """
    K = 1 << (k - 1)
    zero = parent(L[0])(0)
    zeroes = [zero] * K
    return [[L[i + j] for j in range(K)] + zeroes for i in range(0, K << m, K)]


def _combine(L, m, k):
    r"""
    Assumes L is a list of length `2^m`, each entry a list of
    length `2^k`. Combines together into a single list,
    effectively inverting ``_split()``, but overlaying
    coefficients, i.e. list #i gets added in starting at position
    `2^{k-1} i`. Note that the second half of the last list is
    ignored.
    """
    M = 1 << m
    half_K = 1 << (k - 1)
    return [L[0][j] for j in range(half_K)] + \
        [L[i + 1][j] + L[i][j + half_K]
         for i in range(M - 1) for j in range(half_K)]


def _nega_combine(L, m, k):
    r"""
    Same as ``_combine()``, but doesn't ignore the second
    half of the last list; instead it makes that piece wrap around
    negacyclically.
    """
    M = 1 << m
    half_K = 1 << (k - 1)
    return [L[0][j] - L[M - 1][j + half_K] for j in range(half_K)] + \
        [L[i + 1][j] + L[i][j + half_K]
         for i in range(M - 1) for j in range(half_K)]


# -------------------------------------------------------------------
#      FFT-based convolution and negacyclic convolutions


def _negaconvolution(L1, L2, n):
    """
    Negacyclic convolution of L1 and L2.

    L1 and L2 must both have length `2^n`.
    """
    if n <= 3:    # arbitrary cutoff
        return _negaconvolution_naive(L1, L2)
    return _negaconvolution_fft(L1, L2, n)


def _negaconvolution_fft(L1, L2, n):
    r"""
    Return negacyclic convolution of lists L1 and L2, using FFT algorithm.

    L1 and L2 must both have length `2^n`, where `n \geq 3`.
    Assumes all entries of L1 and L2 belong to the same ring.

    EXAMPLES::

        sage: from sage.rings.polynomial.convolution import _negaconvolution_naive
        sage: from sage.rings.polynomial.convolution import _negaconvolution_fft
        sage: _negaconvolution_naive(list(range(8)), list(range(5, 13)))
        [-224, -234, -224, -192, -136, -54, 56, 196]
        sage: _negaconvolution_fft(list(range(8)), list(range(5, 13)), 3)
        [-224, -234, -224, -192, -136, -54, 56, 196]

    TESTS::

        sage: for n in range(3, 10):
        ....:    L1 = [ZZ.random_element(100) for _ in range(1 << n)]
        ....:    L2 = [ZZ.random_element(100) for _ in range(1 << n)]
        ....:    assert _negaconvolution_naive(L1, L2) == _negaconvolution_fft(L1, L2, n)
    """
    assert n >= 3

    R = parent(L1[0])

    # split into 2^m pieces of 2^(k-1) coefficients each, with k as small
    # as possible, subject to m <= k (so that the ring of Fourier coefficients
    # has enough roots of unity)
    m = (n + 1) >> 1
    k = n + 1 - m

    M = 1 << m
    K = 1 << k

    # split inputs into polynomials
    L1 = _split(L1, m, k)
    L2 = _split(L2, m, k)

    # fft each input
    _fft(L1, K, 0, m, K >> 1)
    _fft(L2, K, 0, m, K >> 1)

    # pointwise multiply
    L3 = [_negaconvolution(L1[i], L2[i], k) for i in range(M)]

    # inverse fft
    _ifft(L3, K, 0, m, K >> 1)

    # combine back into a single list
    L3 = _nega_combine(L3, m, k)

    # normalise
    return [R(x / M) for x in L3]


def _convolution_fft(L1, L2):
    r"""
    Return convolution of non-empty lists L1 and L2, using FFT algorithm.

    L1 and L2 may have arbitrary lengths `\geq 4`.
    Assumes all entries of L1 and L2 belong to the same ring.

    EXAMPLES::

        sage: from sage.rings.polynomial.convolution import _convolution_naive
        sage: from sage.rings.polynomial.convolution import _convolution_fft
        sage: _convolution_naive([1, 2, 3], [4, 5, 6])
        [4, 13, 28, 27, 18]
        sage: _convolution_fft([1, 2, 3], [4, 5, 6])
        [4, 13, 28, 27, 18]

    TESTS::

        sage: for len1 in range(4, 30):
        ....:    for len2 in range(4, 30):
        ....:       L1 = [ZZ.random_element(100) for _ in range(len1)]
        ....:       L2 = [ZZ.random_element(100) for _ in range(len2)]
        ....:       assert _convolution_naive(L1, L2) == _convolution_fft(L1, L2)
    """
    R = parent(L1[0])

    # choose n so that output convolution length is 2^n
    len1 = len(L1)
    len2 = len(L2)
    outlen = len1 + len2 - 1
    n = int(ceil(log(outlen) / log(2.0)))

    # split into 2^m pieces of 2^(k-1) coefficients each, with k as small
    # as possible, subject to m <= k + 1 (so that the ring of Fourier
    # coefficients has enough roots of unity)
    m = (n >> 1) + 1
    k = n + 1 - m

    N = 1 << n
    M = 1 << m
    K = 1 << k

    # zero pad inputs up to length N
    zero = R(0)
    L1 = L1 + [zero] * (N - len(L1))
    L2 = L2 + [zero] * (N - len(L2))

    # split inputs into polynomials
    L1 = _split(L1, m, k)
    L2 = _split(L2, m, k)

    # fft each input
    _fft(L1, K, 0, m, K)
    _fft(L2, K, 0, m, K)

    # pointwise multiply
    L3 = [_negaconvolution(L1[i], L2[i], k) for i in range(M)]

    # inverse fft
    _ifft(L3, K, 0, m, K)

    # combine back into a single list
    L3 = _combine(L3, m, k)

    # normalise, and truncate to correct length
    return [R(L3[i] / M) for i in range(outlen)]



# sage_setup: distribution = sagemath-pari
# sage.doctest: needs sage.libs.flint sage.libs.pari sage.modules

from cysignals.signals cimport sig_check, sig_on, sig_str, sig_off

from sage.libs.flint.fmpz cimport *
from sage.libs.flint.fmpz_mat cimport *
from sage.libs.gmp.mpz cimport *
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.rings.integer cimport Integer
from sage.rings.integer_ring import ZZ

# PARI C library
from cypari2.gen cimport Gen
from cypari2.stack cimport clear_stack, new_gen
from cypari2.paridecl cimport *
from sage.libs.pari import pari
from sage.libs.pari.convert_gmp cimport INT_to_mpz
from sage.libs.pari.convert_flint cimport (_new_GEN_from_fmpz_mat_t,
           _new_GEN_from_fmpz_mat_t_rotate90, integer_matrix)
from sage.libs.pari.convert_sage_matrix import gen_to_sage_matrix


def _pari(Matrix_integer_dense self):
    return integer_matrix(self._matrix, 0)


def _det_pari(Matrix_integer_dense self, int flag=0):
    """
    Determinant of this matrix using Gauss-Bareiss. If (optional)
    flag is set to 1, use classical Gaussian elimination.

    For efficiency purposes, this det is computed entirely on the
    PARI stack then the PARI stack is cleared. This function is
    most useful for very small matrices.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_pari import _det_pari
        sage: _det_pari(matrix(ZZ, 0))
        1
        sage: _det_pari(matrix(ZZ, 0), 1)
        1
        sage: _det_pari(matrix(ZZ, 3, [1..9]))
        0
        sage: _det_pari(matrix(ZZ, 3, [1..9]), 1)
        0
    """
    sig_on()
    cdef GEN d = det0(pari_GEN(self), flag)
    # now convert d to a Sage integer e
    cdef Integer e = Integer.__new__(Integer)
    INT_to_mpz(e.value, d)
    clear_stack()
    return e


def _lll_pari(Matrix_integer_dense self):
    # call pari with flag=4: kernel+image
    # pari uses column convention: need to transpose the matrices
    A = integer_matrix(self._matrix, 1)
    K, T = A.qflll(4)
    r = ZZ(T.length())
    # TODO: there is no optimized matrix converter pari -> sage
    U = gen_to_sage_matrix(pari.matconcat([K, T]).mattranspose())
    R = U * self
    return R, U, r


def _rank_pari(Matrix_integer_dense self):
    """
    Rank of this matrix, computed using PARI.  The computation is
    done entirely on the PARI stack, then the PARI stack is
    cleared.  This function is most useful for very small
    matrices.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_pari import _rank_pari
        sage: _rank_pari(matrix(ZZ,3,[1..9]))
        2
    """
    sig_on()
    cdef long r = rank(pari_GEN(self))
    clear_stack()
    return r


def _hnf_pari(Matrix_integer_dense self, int flag=0, bint include_zero_rows=True):
    """
    Hermite normal form of this matrix, computed using PARI.

    INPUT:

    - ``flag`` -- 0 (default), 1, 3 or 4 (see docstring for
      ``pari.mathnf``)

    - ``include_zero_rows`` -- boolean; if ``False``, do not include
      any of the zero rows at the bottom of the matrix in the output

    .. NOTE::

        In no cases is the transformation matrix returned by this
        function.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_pari import _hnf_pari
        sage: _hnf_pari(matrix(ZZ,3,[1..9]))
        [1 2 3]
        [0 3 6]
        [0 0 0]
        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 1)
        [1 2 3]
        [0 3 6]
        [0 0 0]
        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 3)
        [1 2 3]
        [0 3 6]
        [0 0 0]
        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 4)
        [1 2 3]
        [0 3 6]
        [0 0 0]

    Check that ``include_zero_rows=False`` works correctly::

        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 0, include_zero_rows=False)
        [1 2 3]
        [0 3 6]
        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 1, include_zero_rows=False)
        [1 2 3]
        [0 3 6]
        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 3, include_zero_rows=False)
        [1 2 3]
        [0 3 6]
        sage: _hnf_pari(matrix(ZZ,3,[1..9]), 4, include_zero_rows=False)
        [1 2 3]
        [0 3 6]

    Check that :issue:`12346` is fixed::

        sage: pari('mathnf(Mat([0,1]), 4)')
        [Mat(1), [1, 0; 0, 1]]
    """
    sig_on()
    A = _new_GEN_from_fmpz_mat_t_rotate90(self._matrix)
    H = mathnf0(A, flag)
    if typ(H) == t_VEC:
        H = gel(H, 1)
    GenH = new_gen(H)
    return extract_hnf_from_pari_matrix(self, GenH, include_zero_rows)


cdef inline GEN pari_GEN(Matrix_integer_dense B) noexcept:
    r"""
    Create the PARI GEN object on the stack defined by the integer
    matrix B. This is used internally by the function for conversion
    of matrices to PARI.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN A = _new_GEN_from_fmpz_mat_t(B._matrix)
    return A


cdef extract_hnf_from_pari_matrix(Matrix_integer_dense self, Gen H, bint include_zero_rows):
    cdef mpz_t tmp
    mpz_init(tmp)

    # Figure out how many columns we got back.
    cdef long H_nc = glength(H.g)  # number of columns
    # Now get the resulting Hermite form matrix back to Sage, suitably re-arranged.
    cdef Matrix_integer_dense B
    if include_zero_rows:
        B = self.new_matrix()
    else:
        B = self.new_matrix(nrows=H_nc)
    cdef long i, j
    for i in range(self._ncols):
        for j in range(H_nc):
            sig_check()
            INT_to_mpz(tmp, gcoeff(H.g, i+1, H_nc-j))
            fmpz_set_mpz(fmpz_mat_entry(B._matrix,j,self._ncols-i-1),tmp)
    mpz_clear(tmp)
    return B
